#define _GNU_SOURCE

#include <arpa/inet.h>
#include <errno.h>
#include <error.h>
#include <fcntl.h>
#include <getopt.h>
#include <ifaddrs.h>
#include <inttypes.h>
#include <limits.h>
#include <linux/errqueue.h>
#include <linux/ethtool.h>
#include <linux/net_tstamp.h>
#include <linux/sockios.h>
#include <net/if.h>
#include <netdb.h>
#include <netinet/in.h>
#include <poll.h>
#include <pthread.h>
#include <sched.h>
#include <signal.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>
#include <bpf/bpf.h>
#include <bpf/libbpf.h>
#include <bpf/xsk.h>
#include <linux/if_ether.h>
#include <linux/if_link.h>
#include <linux/if_xdp.h>
#include <linux/ip.h>
#include <linux/udp.h>

#define err(...)                  \
	do {                            \
		fprintf(stderr, __VA_ARGS__); \
		exit(EXIT_FAILURE);           \
	} while (0)
#define err_errno(...) error(EXIT_FAILURE, errno, __VA_ARGS__);

#define ETH_FCS_SIZE 4
#define FRAME_SIZE XSK_UMEM__DEFAULT_FRAME_SIZE
#define PACKET_SIZE 262

#define BUFFER_SIZE 4096

#define NB_PACKETS 2000000
#define BATCH_SIZE 256

//#define PRINT_PROGRESS
#define PRINT_PROGRESS_INTERVAL 1000

#define ACTIVE_POLLING

//#define DEBUG
static void log_info(const char * section, const char * msg, ...) {
  time_t t;
  struct tm ts;
	char line[256];
  va_list arglist;

  time(&t);
  ts = *localtime(&t);
  strftime(line, 80, "%m-%d %H:%M:%S", &ts);
  sprintf(line + strlen(line), " INFO [%s] ", section);
  va_start(arglist, msg);
  vsprintf(line + strlen(line), msg, arglist);
  va_end(arglist);
  puts(line);
}
#ifdef DEBUG
static void log_debug(const char * section, const char * msg, ...) {
  time_t t;
  struct tm ts;
	char line[256];
  va_list arglist;

  time(&t);
  ts = *localtime(&t);
  strftime(line, 80, "%m-%d %H:%M:%S", &ts);
  sprintf(line + strlen(line), " DEBUG [%s] ", section);
  va_start(arglist, msg);
  vsprintf(line + strlen(line), msg, arglist);
  va_end(arglist);
  puts(line);
}
#else
#define log_debug(...)
#endif
struct xsk_umem_info {
	struct xsk_ring_prod fq;
	struct xsk_ring_cons cq;
	struct xsk_umem *umem;
	void *buffer;
};
struct xdpsock {
	struct xsk_ring_cons rx;
	struct xsk_ring_prod tx;
	struct xsk_umem_info umem;
	struct xsk_socket *xsk;
	int fd;
};
// Timestamps utils
#define NSEC_PER_SEC INT64_C(1000000000)
static struct timespec int_to_ts(int64_t t) {
  struct timespec ts;
  ts.tv_sec = t / NSEC_PER_SEC;
  ts.tv_nsec = t - (ts.tv_sec * NSEC_PER_SEC);
  return ts;
}
static int64_t ts_to_int(struct timespec ts) {
    return ts.tv_sec * NSEC_PER_SEC + ts.tv_nsec;
}
static void add_ns(struct timespec *t, int64_t ns) {
  t->tv_nsec += ns;
  while (t->tv_nsec >= ((int64_t)NSEC_PER_SEC)) {
    t->tv_sec += 1;
    t->tv_nsec -= NSEC_PER_SEC;
  }
}
static int64_t calcdiff_ns(struct timespec t1, struct timespec t2) {
  int64_t diff;
  diff = NSEC_PER_SEC * ((int)t1.tv_sec - (int)t2.tv_sec);
  diff += ((int)t1.tv_nsec - (int)t2.tv_nsec);
  return diff;
}

static void init_xdp_recv(char * network_if);
static void init_xdp_send(char * network_if);
static int recv_xdp_packet(void);
static void init_recv(void);
static void send_xdp_packet(void);
static void recv_xdp_cleanup(void);
static void close_xdp_socket(void);

static uint8_t pkt_data[FRAME_SIZE];
static struct xdpsock send_xdp_socket;
static struct xdpsock recv_xdp_socket;
static int received;
static int64_t sent;
static uint32_t idx_rx = 0, idx_recv;
static int batch_size = BATCH_SIZE;
int trace_fd;

int main() {
  char * network_if = "ens5f0np0";

  log_info("", "Starting");

  init_xdp_send(network_if);
  //init_xdp_recv(network_if);
  return 0;
}

void trace_write(const char *fmt, ...)
{
    va_list ap;
    char buf[256];
    int n;

    if (trace_fd < 0)
        return;
    va_start(ap, fmt);
    n = vsnprintf(buf, 256, fmt, ap);
    va_end(ap);
    write(trace_fd, buf, n);
}

static int latency_target_fd = -1;
static int32_t latency_target_value = 0;
void set_latency_target(void) {
  struct stat s;
  int err;

  errno = 0;
  err = stat("/dev/cpu_dma_latency", &s);
  if (err == -1) {
    error(EXIT_FAILURE, errno, "WARN: stat /dev/cpu_dma_latency failed");
    return;
  }

  errno = 0;
  latency_target_fd = open("/dev/cpu_dma_latency", O_RDWR);
  if (latency_target_fd == -1) {
    error(EXIT_FAILURE, errno, "WARN: open /dev/cpu_dma_latency");
    return;
  }

  errno = 0;
  err = write(latency_target_fd, &latency_target_value, 4);
  if (err < 1) {
    error(EXIT_FAILURE, errno, "# error setting cpu_dma_latency to %d!",
          latency_target_value);
    close(latency_target_fd);
    return;
  }
  printf("# /dev/cpu_dma_latency set to %dus\n", latency_target_value);
}
static void init_xdp_send(char * network_if) {
  uint32_t idx;
  int ret, i, xsks_map = 0;
  struct xsk_socket_config xsk_cfg;
  struct xsk_umem_config cfg = {
      .fill_size = XSK_RING_PROD__DEFAULT_NUM_DESCS,
      .comp_size = XSK_RING_CONS__DEFAULT_NUM_DESCS,
      .frame_size = FRAME_SIZE,
      .frame_headroom = XSK_UMEM__DEFAULT_FRAME_HEADROOM,
      .flags = 0,
  };
  void *buffer = NULL;
  struct ethhdr *eth_hdr = (struct ethhdr *)pkt_data;

	//set_latency_target();
  if (mlockall(MCL_CURRENT | MCL_FUTURE)) {
    fprintf(stderr, "mlockall failed");
    exit(EXIT_FAILURE);
  }

	for(int j = 0; j < PACKET_SIZE; j++)
		pkt_data[j] = 0x17;

  memcpy(eth_hdr->h_dest, "\xb8\x59\x9f\x07\x7d\xdb", ETH_ALEN);
  memcpy(eth_hdr->h_source, "\x04\x09\xa5\x0f\x9f\x4c", ETH_ALEN);
  eth_hdr->h_proto = htons(ETH_P_IP);

	//for(int j = 0; j < PACKET_SIZE; j++)
	//	printf("%x", pkt_data[j]);
	//printf("\n");
	//exit(EXIT_SUCCESS);

  log_debug("", "posix_memalign");
  /* Allocate user space memory for xdp frames */
  ret = posix_memalign(&buffer, sysconf(_SC_PAGE_SIZE), BUFFER_SIZE * FRAME_SIZE);
  if (ret) err_errno("posix_memalign() failed");

  log_debug("", "xsk_umem__create");
  ret = xsk_umem__create(&send_xdp_socket.umem.umem, buffer, BUFFER_SIZE * FRAME_SIZE,
                         &send_xdp_socket.umem.fq, &send_xdp_socket.umem.cq, &cfg);
  if (ret) err("xsk_umem__create() failed");
  send_xdp_socket.umem.buffer = buffer;

  log_debug("", "open_xdp_socket");

  /* Create XDP socket */
  xsk_cfg.rx_size = XSK_RING_CONS__DEFAULT_NUM_DESCS;
  xsk_cfg.tx_size = XSK_RING_PROD__DEFAULT_NUM_DESCS;
  xsk_cfg.libbpf_flags = 0;
  xsk_cfg.xdp_flags = XDP_FLAGS_DRV_MODE;
  xsk_cfg.bind_flags = XDP_USE_NEED_WAKEUP;

  log_debug("", "xsk_socket__create");
  ret = xsk_socket__create(&send_xdp_socket.xsk, network_if, 0, send_xdp_socket.umem.umem,
                           &send_xdp_socket.rx, &send_xdp_socket.tx, &xsk_cfg);

  log_debug("", "send_xdp_packet");
  send_xdp_packet();
}
static void init_xdp_recv(char * network_if) {
  uint32_t idx;
  int ret, i, xsks_map = 0;
  struct xsk_socket_config xsk_cfg;
  struct xsk_umem_config cfg = {
      .fill_size = XSK_RING_PROD__DEFAULT_NUM_DESCS,
      .comp_size = XSK_RING_CONS__DEFAULT_NUM_DESCS,
      .frame_size = FRAME_SIZE,
      .frame_headroom = XSK_UMEM__DEFAULT_FRAME_HEADROOM,
      .flags = 0,
  };
  void *buffer = NULL;
  sent = 0;

  log_debug("", "posix_memalign");
  /* Allocate user space memory for xdp frames */
  ret = posix_memalign(&buffer, sysconf(_SC_PAGE_SIZE), BUFFER_SIZE * FRAME_SIZE);
  if (ret) err_errno("posix_memalign() failed");

  log_debug("", "xsk_umem__create");
  ret = xsk_umem__create(&recv_xdp_socket.umem.umem, buffer, BUFFER_SIZE * FRAME_SIZE,
                         &recv_xdp_socket.umem.fq, &recv_xdp_socket.umem.cq, &cfg);
  if (ret) err("xsk_umem__create() failed");
  recv_xdp_socket.umem.buffer = buffer;

  log_debug("", "open_xdp_socket");

  /* Create XDP socket */
  xsk_cfg.rx_size = XSK_RING_CONS__DEFAULT_NUM_DESCS;
  xsk_cfg.tx_size = XSK_RING_PROD__DEFAULT_NUM_DESCS;
  xsk_cfg.libbpf_flags = 0;
  xsk_cfg.xdp_flags = XDP_FLAGS_DRV_MODE;
  xsk_cfg.bind_flags = 0;

  log_debug("", "xsk_socket__create");
  ret = xsk_socket__create(&recv_xdp_socket.xsk, network_if, 0, recv_xdp_socket.umem.umem,
                           &recv_xdp_socket.rx, &recv_xdp_socket.tx, &xsk_cfg);

  log_debug("", "recv_xdp_packet");
  recv_xdp_packet();
}
static void complete_tx_only(void) {
  int rcvd;
  uint32_t idx;
  rcvd = xsk_ring_cons__peek(&send_xdp_socket.umem.cq, batch_size, &idx);
  if (rcvd > 0) {
    xsk_ring_cons__release(&send_xdp_socket.umem.cq, rcvd);
    sent += rcvd;
  }
}

static void send_xdp_packet(void) {
  struct pollfd fds[1] = {};
  int i, ret;
  struct timespec start, end;
  int64_t duration_ns;
#ifdef PRINT_PROGRESS
	int64_t progress = PRINT_PROGRESS_INTERVAL;
#endif
  fds[0].fd = xsk_socket__fd(send_xdp_socket.xsk);
  fds[0].events = POLLOUT;

  for (int i = 0; i < BUFFER_SIZE; i++)
    memcpy(xsk_umem__get_data(send_xdp_socket.umem.buffer, i * FRAME_SIZE), pkt_data, PACKET_SIZE - ETH_FCS_SIZE);

  clock_gettime(CLOCK_MONOTONIC, &start);
  for(int l = 0; (sent < NB_PACKETS); l++) {
    uint32_t idx;

#ifdef PRINT_PROGRESS
		if(sent > progress) {
			printf("%" PRIi64 "\n", sent);
			progress += PRINT_PROGRESS_INTERVAL;
		}
#endif
#ifdef ACTIVE_POLLING
    ret = poll(fds, 1, 0);
    if ((ret <= 0) || !(fds[0].revents & POLLOUT)) {
			complete_tx_only();
      continue;
		}
#else
    ret = poll(fds, 1, 100);
    if ((ret <= 0) || !(fds[0].revents & POLLOUT)) {
			complete_tx_only();
      continue;
		}
#endif

    while(xsk_ring_prod__reserve(&send_xdp_socket.tx, batch_size, &idx) < batch_size)
      complete_tx_only();

    for (int k = 0; k < batch_size; k++) {
      struct xdp_desc *tx_desc = xsk_ring_prod__tx_desc(&send_xdp_socket.tx, idx + k);
      tx_desc->addr = k * (FRAME_SIZE);
      tx_desc->len = PACKET_SIZE - ETH_FCS_SIZE;
    }

    xsk_ring_prod__submit(&send_xdp_socket.tx, batch_size);
    if (xsk_ring_prod__needs_wakeup(&send_xdp_socket.tx))
          sendto(xsk_socket__fd(send_xdp_socket.xsk), NULL, 0, MSG_DONTWAIT, NULL, 0);
  }
  clock_gettime(CLOCK_MONOTONIC, &end);
  duration_ns = calcdiff_ns(end, start);
  log_info("STATS", "Duration: %" PRIi64 " us", duration_ns / 1000);
  log_info("STATS", "Packets sent: %" PRIi64 " / %" PRIi64, sent, NB_PACKETS);
  log_info("STATS", "pps: %" PRIi64, (sent * NSEC_PER_SEC) / duration_ns);
}

static int recv_xdp_packet(void) {
  int ret;
  static struct pollfd fds_recv[1] = {0};

  fds_recv[0].fd = xsk_socket__fd(recv_xdp_socket.xsk);
  fds_recv[0].events = POLLIN;
  /* Add some buffers in fill ring */
  log_debug("", "xsk_ring_prod__reserve");
  ret = xsk_ring_prod__reserve(&recv_xdp_socket.umem.fq,
                               XSK_RING_PROD__DEFAULT_NUM_DESCS, &idx_recv);
  if (ret != XSK_RING_PROD__DEFAULT_NUM_DESCS)
    err("xsk_ring_prod__reserve() failed");
  log_debug("", "xsk_ring_prod__fill_addr");
  for (int i = 0; i < XSK_RING_PROD__DEFAULT_NUM_DESCS; i++)
    *xsk_ring_prod__fill_addr(&recv_xdp_socket.umem.fq, idx_recv++) = i * FRAME_SIZE;
  log_debug("", "xsk_ring_prod__submit");
  xsk_ring_prod__submit(&recv_xdp_socket.umem.fq, XSK_RING_PROD__DEFAULT_NUM_DESCS);

  ret = poll(fds_recv, 1, -1);

  if (ret != 1)
    error(EXIT_FAILURE, errno, "poll failed");

  log_info("", "xsk_ring_cons__peek");
  received = xsk_ring_cons__peek(&recv_xdp_socket.rx, 1, &idx_rx);
  log_info("", "Received packet ! (%d)", received);

  if (received != 1)
    error(EXIT_FAILURE, errno, "Received %d packets", received);
  return 0;
}
static void recv_xdp_cleanup(void) {
  uint64_t addr;
  int ret;

  /* Cleanup */
  xsk_ring_cons__release(&recv_xdp_socket.rx, received);

  /* Add that particular buffer back to the fill queue */
  if (xsk_prod_nb_free(&recv_xdp_socket.umem.fq, received)) {
    ret = xsk_ring_prod__reserve(&recv_xdp_socket.umem.fq, received, &idx_recv);

    if (ret != received) err("xsk_ring_prod__reserve() failed");

    *xsk_ring_prod__fill_addr(&recv_xdp_socket.umem.fq, idx_recv) =
        xsk_umem__extract_addr(addr);

    xsk_ring_prod__submit(&recv_xdp_socket.umem.fq, received);
  }
}
static void close_xdp_socket(void) {
  xsk_socket__delete(send_xdp_socket.xsk);
  xsk_socket__delete(recv_xdp_socket.xsk);
  xsk_umem__delete(send_xdp_socket.umem.umem);
  xsk_umem__delete(recv_xdp_socket.umem.umem);
}
